!----------------------------------------------------------
!   resolution de les quations d'euler 2d 
!   - sur grille structure curviligne
!   - domaine topologiquement equivalent a un carre : 
!             NORD
!            _______  
!            |     |  ^j
!    OUEST   |     |  |     EST
!            _______  --> i

!              SUD
!   - schema volumes finis, decentre d'ordre 1, explicite en temps
!   - fichiers de donnees : maillage.data
!----------------------------------------------------------

program euler2d

  !----------- sous programmes ------------------------------------ 
  use mod_precision   !- (double ou simple precision)
  use mod_flux
  !----------------------------------------------------------------

  implicit none

  !------------ variables -----------------------------------------
  integer :: i,j,imax,jmax,nmax,n,k,pas_affichage,i_entree
  character(len=100) :: fich,numero
  real(pr) :: L,h,dx,dy,cfl,invdt,dt,vmax,tmax
  real(pr) :: r_gaz,norme_u,perimetre,ut,un
  real(pr), dimension(:,:), allocatable :: x,y,xc,yc
  real(pr), dimension(:,:), allocatable :: rho,ux,uy,t,e,p,a,b
  real(pr), dimension(:,:), allocatable :: aire,taux,tauy,nux,nuy,l_tau,l_nu
  real(pr), dimension(:,:,:), allocatable :: vect_u,vect_unew,flux_tau,flux_nu
  !----------------------------------------------------------------


  !------------- programme ----------------------------------------

  !--- lecture du maillage
  open(unit=10,file='maillage.data')

  !-nb de points dans la grille
  read(10,*) imax,jmax   

  !-nb de mailles
  imax=imax-1
  jmax=jmax-1   

  allocate(x(0:imax,0:jmax),y(0:imax,0:jmax))
  do i=0,imax
     do j=0,jmax
        read(10,*) x(i,j),y(i,j)
     enddo
  enddo
  close(10)

  !--- calcul des aires des mailles
  allocate(aire(1:imax,1:jmax))
  do i=1,imax
     do j=1,jmax
        aire(i,j)=0.5_pr*( (x(i,j)-x(i-1,j-1))*(y(i-1,j)-y(i,j-1))             &
             &            -(y(i,j)-y(i-1,j-1))*(x(i-1,j)-x(i,j-1)) )
     enddo
  enddo

  !--- calcul des normales unitaires et des longeurs des interfaces
  allocate(taux(0:imax,1:jmax),tauy(0:imax,1:jmax),l_tau(0:imax,1:jmax))
  allocate(nux(1:imax,0:jmax),nuy(1:imax,0:jmax),l_nu(1:imax,0:jmax))
  do i=0,imax
     do j=1,jmax
        taux(i,j) = y(i,j)-y(i,j-1)
        tauy(i,j) = -(x(i,j)-x(i,j-1))
        l_tau(i,j)=sqrt(taux(i,j)**2+tauy(i,j)**2)
        taux(i,j)=taux(i,j)/l_tau(i,j)
        tauy(i,j)=tauy(i,j)/l_tau(i,j)
     end do
  end do
  do i=1,imax
     do j=0,jmax
        nux(i,j) = -(y(i,j)-y(i-1,j))
        nuy(i,j) = x(i,j)-x(i-1,j)
        l_nu(i,j)=sqrt(nux(i,j)**2+nuy(i,j)**2)
        nux(i,j)=nux(i,j)/l_nu(i,j)
        nuy(i,j)=nuy(i,j)/l_nu(i,j)
     end do
  end do

  !--- donne initiale : 
  allocate(rho(0:imax+1,0:jmax+1),ux(0:imax+1,0:jmax+1),uy(0:imax+1,0:jmax+1),t(0:imax+1,0:jmax+1))
  do i=1,imax
     do j=1,jmax

        if (0.25_pr*(x(i,j)+x(i-1,j)+x(i,j-1)+x(i-1,j-1))<x(0,1)) then
           rho(i,j)= 0.2969689477e-4_pr
           ux(i,j) = 1059.458022_pr
           uy(i,j) = 0.0_pr
           t(i,j) = 1295.646765_pr 
        else
           rho(i,j)=0.13e-4_pr
           ux(i,j) =0.0_pr
           uy(i,j) =0.0_pr
           t(i,j) =235.8476445_pr
        end if

     end do
  end do

  !--- conditions aux limites : 
  !-- exemple 1 : 
  rho(0,:)=rho(1,:)      !- ouest (symtrie)
  ux(0,:)=ux(1,:)
  uy(0,:)=-uy(1,:)
  t(0,:)=t(1,:)

  rho(imax+1,:)=rho(imax,:)    !- est (symtrie)
  ux(imax+1,:)=ux(imax,:)
  uy(imax+1,:)=-uy(imax,:)
  t(imax+1,:)=t(imax,:)

  rho(:,0)=rho(:,1)          !- sud (paroi solide)
  t(:,0)=t(:,1)
  do i=1,imax
     un=nux(i,0)*ux(i,1)+nuy(i,0)*uy(i,1)
     ut=-nuy(i,0)*ux(i,1)+nux(i,0)*uy(i,1)
     ux(i,0)=nux(i,0)*(-un)-nuy(i,0)*ut
     uy(i,0)=nuy(i,0)*(-un)+nux(i,0)*ut
  end do

  i_entree=80
  rho(1:i_entree,jmax+1)=0.2969689477e-4_pr     !- nord (entre)
  ux(1:i_entree,jmax+1)=1059.458022_pr
  uy(1:i_entree,jmax+1)=0.0_pr
  t(1:i_entree,jmax+1)=1295.646765_pr 

  rho(i_entree+1:imax,jmax+1)=rho(i_entree+1:imax,jmax)     !- nord (sortie)
  ux(i_entree+1:imax,jmax+1)=ux(i_entree+1:imax,jmax)
  uy(i_entree+1:imax,jmax+1)=uy(i_entree+1:imax,jmax)
  t(i_entree+1:imax,jmax+1)=t(i_entree+1:imax,jmax)

  !--- valeurs de t aux 4 mailles fictives de coin : inutilises dans le schma
  !--- mais donne division par 0 dans le calcul de a
  t(0,0)=1._pr
  t(0,jmax+1)=1._pr
  t(imax+1,jmax+1)=1._pr
  t(imax+1,0)=1._pr

  !--- calcul des quantits drives
  allocate(p(0:imax+1,0:jmax+1),a(0:imax+1,0:jmax+1),b(0:imax+1,0:jmax+1),e(0:imax+1,0:jmax+1))
  r_gaz=1.3806503e-23_pr/.663e-25_pr
  p=rho*r_gaz*t
  b=sqrt(3*r_gaz*t)
  a=rho/(8_pr*b**3)
  e=0.5_pr*rho*(ux**2+uy**2)+3._pr/2._pr*rho*r_gaz*t

  !--- calcul du vecteur des quantits conservatives
  allocate(vect_u(4,1:imax,1:jmax),vect_unew(4,1:imax,1:jmax))
  vect_u(1,:,:)=rho(1:imax,1:jmax)
  vect_u(2,:,:)=rho(1:imax,1:jmax)*ux(1:imax,1:jmax)
  vect_u(3,:,:)=rho(1:imax,1:jmax)*uy(1:imax,1:jmax)
  vect_u(4,:,:)=e(1:imax,1:jmax)
  vect_unew=vect_u

  !--- pas de temps et vitesse maximum
  cfl=0.9_pr
  invdt=0._pr 
  do i=1,imax
     do j=1,jmax
        norme_u=sqrt(ux(i,j)**2+uy(i,j)**2)
        perimetre=l_tau(i-1,j)+l_tau(i,j)+l_nu(i,j-1)+l_nu(i,j)
        invdt=max(invdt,(norme_u+b(i,j)*perimetre/aire(i,j)))
     end do
  end do
  dt=cfl/invdt

  !--- allocation des flux
  allocate(flux_tau(4,0:imax,1:jmax),flux_nu(4,1:imax,0:jmax))

  !--- temps de simulation et parametres de sorties
  tmax=.3828823925e-2_pr
  nmax=floor(tmax/dt) 

  print*,"nombre d'iterations en temps :",nmax
  pas_affichage=max(nmax/120,1)  !-- pour avoir 120 images, soit 5 secondes d'animation
  k=0

  !--- evolution
  do n=1,nmax
     !-- calcul des flux
     do i=0,imax
        do j=1,jmax
           flux_tau(:,i,j) = l_tau(i,j)*(fluxp(rho(i,j),ux(i,j),uy(i,j),e(i,j),p(i,j),t(i,j)               &
                ,a(i,j),b(i,j),taux(i,j),tauy(i,j))                         &
                &                      + fluxm(rho(i+1,j),ux(i+1,j),uy(i+1,j),e(i+1,j),p(i+1,j),t(i+1,j)   &
                &                              ,a(i+1,j),b(i+1,j),taux(i,j),tauy(i,j)) ) 
        end do
     end do

     do i=1,imax
        do j=0,jmax
           flux_nu(:,i,j) = l_nu(i,j)*(fluxp(rho(i,j),ux(i,j),uy(i,j),e(i,j),p(i,j),t(i,j)                 &
                &                            ,a(i,j),b(i,j),nux(i,j),nuy(i,j))                             &
                &                      + fluxm(rho(i,j+1),ux(i,j+1),uy(i,j+1),e(i,j+1),p(i,j+1),t(i,j+1)   &
                &                              ,a(i,j+1),b(i,j+1),nux(i,j),nuy(i,j)) ) 
        end do
     end do


     !-- iteration en temps
     do i=1,imax
        do j=1,jmax
           vect_unew(:,i,j) = vect_u(:,i,j)                                              &
                &             - dt/aire(i,j)*( (flux_tau(:,i,j)-flux_tau(:,i-1,j))       &
                &                               + (flux_nu(:,i,j)-flux_nu(:,i,j-1))      &
                &                            )
        end do
     end do

     !-- mise a jour
     vect_u=vect_unew

     !-- calcul de rho,ux,uy,t
     do i=1,imax
        do j=1,jmax
           rho(i,j)=vect_u(1,i,j)
           ux(i,j)=vect_u(2,i,j)/rho(i,j)
           uy(i,j)=vect_u(3,i,j)/rho(i,j)
           t(i,j)=2._pr/(3._pr*r_gaz*rho(i,j))*(vect_u(4,i,j)-0.5_pr*rho(i,j)*(ux(i,j)**2+uy(i,j)**2))
        end do
     end do


     !-- mise a jour cl
     !-- exemple 1 : symtrie sur O et E, paroi solide au S, sortie sur partie N
     rho(0,1:jmax)=rho(1,1:jmax)      !- ouest (symtrie)
     ux(0,1:jmax)=ux(1,1:jmax)
     uy(0,1:jmax)=-uy(1,1:jmax)
     t(0,1:jmax)=t(1,1:jmax)

     rho(imax+1,1:jmax)=rho(imax,1:jmax)    !- est (symtrie)
     ux(imax+1,1:jmax)=ux(imax,1:jmax)
     uy(imax+1,1:jmax)=-uy(imax,1:jmax)
     t(imax+1,1:jmax)=t(imax,1:jmax)

     rho(1:imax,0)=rho(1:imax,1)          !- sud (paroi solide)
     t(1:imax,0)=t(1:imax,1)
     do i=1,imax
        un=nux(i,0)*ux(i,1)+nuy(i,0)*uy(i,1)
        ut=-nuy(i,0)*ux(i,1)+nux(i,0)*uy(i,1)
        ux(i,0)=nux(i,0)*(-un)-nuy(i,0)*ut
        uy(i,0)=nuy(i,0)*(-un)+nux(i,0)*ut
     end do

     rho(i_entree+1:imax,jmax+1)=rho(i_entree+1:imax,jmax)     !- nord (sortie)
     ux(i_entree+1:imax,jmax+1)=ux(i_entree+1:imax,jmax)
     uy(i_entree+1:imax,jmax+1)=uy(i_entree+1:imax,jmax)
     t(i_entree+1:imax,jmax+1)=t(i_entree+1:imax,jmax)

     !-- mise  jour des quantits drives (y compris dans les mailles fictives)
     p=rho*r_gaz*t
     do i=0,imax+1
        do j=0,jmax+1
           if (t(i,j)<=0._pr) then
              print*,'t=',t(i,j),'en i,j=',i,j
           end if
        end do
     end do
     b=sqrt(3*r_gaz*t)
     a=rho/(8_pr*b**3)
     e=0.5_pr*rho*(ux**2+uy**2)+3._pr/2._pr*rho*r_gaz*t

     !-- sauvegarde resultats (format vtk, lisible par visit)
     if (mod(n,pas_affichage)==0) then

        k=k+1
        write(numero,*) k
        if (k<=9) then
           numero='00'//trim(adjustl(numero))
        elseif (9<k.and.k<=99) then
           numero='0'//trim(adjustl(numero))
        end if

        fich='SORTIES/resultats.'//trim(adjustl(numero))//'.vtk'
        open(unit=1,file=fich)
        write(1,'(1A26)') '# vtk DataFile Version 2.0'
        write(1,*) 'convection 2D'
        write(1,*) 'ASCII'
        write(1,*) 'DATASET STRUCTURED_GRID'
        write(1,*) 'DIMENSIONS ',imax+1,jmax+1,1
        write(1,*) 'POINTS',(imax+1)*(jmax+1),' double'
        do j=0,jmax
           do i=0,imax
              write(1,*) x(i,j),y(i,j),0.0_pr
           end do
        end do
        write(1,*) 'CELL_DATA',imax*jmax
        write(1,*) 'SCALARS rho double'
        write(1,*) 'LOOKUP_TABLE default'
        do j=1,jmax
           do i=1,imax
              write(1,*) rho(i,j)
           end do
        end do
        write(1,*) 'SCALARS T double'
        write(1,*) 'LOOKUP_TABLE default'
        do j=1,jmax
           do i=1,imax
              write(1,*) t(i,j)
           end do
        end do
        write(1,*) 'SCALARS p double'
        write(1,*) 'LOOKUP_TABLE default'
        do j=1,jmax
           do i=1,imax
              write(1,*) p(i,j)
           end do
        end do
        write(1,*) 'VECTORS u double'
        do j=1,jmax
           do i=1,imax
              write(1,*) ux(i,j),uy(i,j),0.0_pr
           end do
        end do
        close(1)

        print*,n

     end if

  end do

  !----------------------------------------------------------------

end program euler2d
